#pragma once

#include <cctype> // for toupper
#include <algorithm>
#include <string>
#include <list>
#include <atlcoll.h>
#include <queue>
#include <direct.h>
#include <math.h>
#include <process.h>

#define INFINITY			1e30
#define MIN_OCC				3
#define NUM_FOLDS			10

#define NEIGHBOUR_COUNTING	1
#define CHI_SQUARE			2
#define FS_WEIGHT			3
#define FUNCTIONALFLOW		4
#define NAIVE_BAYESIAN		5

#define CD_DIST				1
#define FS_SIM				2
#define FS_SIM_R			3
#define PATH_RATIO			4
#define HYPER_GEO			5
#define GEOMETRIC			6
#define CORRELATION			7

#define FACTORIAL_MAX		1000000

using namespace std;


double log_factorial(int i);
double combinatorial(int N, int k);
double factorial(int i);
double log_combinatorial(int N, int k);

class FC;
class ORF;
class NeighbourRelation;

typedef CRBMap <short, FC*> FCMap;
typedef CRBMap <string, ORF*> AnnotMap;
typedef list <FC*> FCList; 
typedef list <ORF*> AnnotList; 
typedef CRBMap <ORF*, NeighbourRelation*> NeighbourList; 
typedef CRBMap <ORF*, double> NeighbourMap; 

extern double g_lambda;
extern double g_lambda_r;
extern double g_reliability;

class PredictScore
{
public:
	ORF* _orf;
	FC* _fc;
	double _score;

	PredictScore(ORF* orf, FC *fc, double score)
	{
		_orf = orf;
		_fc = fc;
		_score = score;
	}

	int operator < (const PredictScore &a)
	{
		return _score < a._score;
	}
};
typedef list <PredictScore> PredictScoreList;


class Score
{
public:
	double _score;
	bool _correct;
//	FC *_fc;
	
	Score(double score, bool correct)
	{
		_score = score;
		_correct = correct;
	}

	int operator < (const Score &a)
	{
		return _score < a._score;
	}
};
typedef list <Score> ResultList;

class FCScore
{
public:
	FC* _fc;
	double _score;

	FCScore(FC *fc, double score)
	{
		_fc = fc;
		_score = score;
	}

	int operator < (const FCScore &a)
	{
		return _score < a._score;
	}
};
typedef list <FCScore> ScoreList;

class ORFScore
{
public:
	ORF* _orf;
	double _score;

	ORFScore(ORF *orf, double score)
	{
		_orf = orf;
		_score = score;
	}

	int operator < (const ORFScore &a)
	{
		return _score < a._score;
	}
};
typedef list <ORFScore> AnnotScoreList;

class NeighbourRelation
{
public:
	bool _done;
	double _weight, _weight2;
	CRBMap <string, int> _types;

	NeighbourRelation(string &type, double weight = 1)
	{
		_weight = 1;
		_weight2 = 1;
		_types.SetAt(type, 1);
	}

	NeighbourRelation(const NeighbourRelation &relation)
	{
		_weight = relation._weight;
		_weight2 = relation._weight2;

		string key;
		int count;
		POSITION pos = relation._types.GetHeadPosition();
		while(pos)
		{
			relation._types.GetNextAssoc(pos, key, count);
			_types.SetAt(key, count);
		}
	}

	int getObservationCount()
	{
		int observations = 0;
		POSITION pos = _types.GetHeadPosition();
		while(pos)
			observations += _types.GetNextValue(pos);
		return observations;
	}

	int getObservationTypeCount()
	{
		return _types.GetCount();
	}
};

class FC
{
public:
	short _id;				// ID
	string _name;			// Name of Functional Class
	FC *_parent;			// Parent of Node
	FCMap _children;		// Children of Node
	AnnotMap _orfs;		// ORFs with this function

	double _occ, _occ_int, _backupocc, _roc;
	double _score;

	FC(short id, string &name, FC* parent);
	~FC();

	string getFC();
	void removeORF(ORF* annot);
	bool haveORF(ORF* annot);
};

class ORF
{
public:
	string _name;
	FCList _fc, _newfc, _ofc,_backupfc;
	CRBMap<FC*, double> _probs, _probs_new;

	ScoreList _predictions;
	NeighbourList _neighbours;

	double _score;
	int _component;

	// Constructor
	ORF(string &name, FC *fc = NULL);
	~ORF();

	// operations
	int addNeighbour(ORF* annot, string &type, double weight=0);
	void removeNeighbour(ORF* annot);
	void removeFC(FC* fc);
	double computeCWeight(ORF *annot);

	double computeCD_dist_OLD(ORF *annot);
	double computeCD_sim(ORF *annot, bool reliability = false);
	double computeFS_weight(ORF *annot, int N, bool reliability = false);
	double computeHyperGeometricScore(ORF *annot, int N);
	double computeGeometric(ORF *annot, bool reliability = false);

	// properties
	bool hasFC(FC* fc);
	bool hasOFC(FC* fc);
	bool hasBackupFC(FC* fc);
	bool hasNeighbour(ORF* annot);
	int getFCCount();

	int getAnnotatedL1NeighbourCount();
	int getAnnotatedOL1NeighbourCount();
	int getAnnotatedL2NeighbourCount();

	int getSharedFCcount(ORF* annot);
	double getSharedFCWeight(ORF* annot);
	double getSharedL2FCWeight(ORF* annot);

	double getNeighbourFCCount();
	double getNeighbourSharedFCWeight(ORF* annot);
	double getSharedNeighbourWeight(ORF* annot);

};

class TreeNode;

typedef list <TreeNode*> TreeNodeList;

class TreeNode
{
public:
	TreeNode *_parent;
	TreeNodeList _children;
	ORF *_annot;

	TreeNode(TreeNode *parent, ORF *annot)
	{
		_parent = parent;
		_annot = annot;
	}
};

class AnnotTree
{
private:
	TreeNode *_root;

	TreeNode* _readSubTree(ifstream &treefile, AnnotMap &map, TreeNode *parent);
	void _writeSubTree(ofstream &treefile, TreeNode *subtree);
	TreeNode* _findNode(ORF *annot, TreeNode* subtree);
	void _getAnnots(TreeNode* subtree, AnnotList &list);

public:

	AnnotTree()
	{
		_root = NULL;
	}

	void readTree(char *filename, AnnotMap &map);
	void writeTree(char *filename);
	ScoreList predict(ORF *annot);
};

typedef CRBMap <FC*, double> FCWeightMap;

class FCManager
{
protected:
	FCMap _FCs;
	AnnotMap _activeORFs;
	FCList _informativeFCs;
	CRBMap <string, int> _typeocc;
	CRBMap <string, double> _typeprob;
	int _interactions;

	int _occ_base, _occ_int_base;
	double _maxweight;

	void _InsertFC(string &funcat, string &name);
	FC* _FindFC(string &funcat);

	void _PrintFCs(ofstream &ofile, string prefix, FC *fc=NULL);
	void _getFCCount(int &count, FC *fc=NULL);

	int _FilterUnInformative(int delta, FC *fc=NULL);
	void _initCrossValidation(FC* fc, int splits, int index);
	
	int _Flood(ORF *annot);
	void _FilterMinorComponents();

public:
	FCManager();
	FCManager(const char *schemefile, const char *annotfile, const char *gridfile);
	~FCManager(void);

	int GetAnnotCount();
	int GetInteractionCount();
	int GetAnnotated();

	inline int GetFCCount()
	{
		int count = 0;
		_getFCCount(count);
		return count;
	}

	inline int GetOrfCount()
	{
		return _activeORFs.GetCount();
	}

	// Zhou et al 2002 - informative FC
	void RemoveUnInformative(int delta = 30);

	// Outputs
	void PrintFCs(const char *filename, int delta = 30);
	void PrintAnnotations(const char *filename);
	void PrintNeighbours(const char *filename);
	void PrintORFS(const char *filename);
	void PrintStatistics(const char *filename);

	// Validation
	void prepareProbs();
	void PrintCrossValidate(ofstream &ofile, short method, bool LOO, double significance = 0.9);
	void PrintUnknownPredicions(ofstream &ofile, short method, double precision);

	// PRODISTIN
	void PrintProdistinMatrix();
	void PrintProdistin();

	// Dataset Manipulation
	void FilterIsolated();
	void RemoveFunctions(int num_fold, int fold);
	void RestoreFunctions();

	// Approaches
	void Do_NeighbourCounting(ORF *annot);
	void Do_Chi_Square(ORF *annot);
	void Do_FunctionalFlow(ORF *annot);
	void Do_FS_Weight(ORF *annot, bool background);
	void Do_Naive_Bayesian(ORF *annot);

};
